#############################################################################
## Miscellaneous utility methods.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: generalutils.rb,v 1.11 2010/08/09 23:41:06 youngh Exp $
#############################################################################

module ArkUtil

  # {period} should be in seconds.
  def self.human_time_period(period)
    return "0 hours" if period < 1800  # including negatives
    days = period / 86400
    hours = ((period % 86400) / 3600.0).round.to_int
    result = []
    result << count_with_unit(days, "day") if days > 0
    result << count_with_unit(hours, "hour") if hours > 0
    result.join(" ")
  end


  # Returns the timestamp for the beginning (00:00:00) of today.
  def self.beginning_of_today_utc
    t = Time.now.to_i.truncate
    t - t % 86400
  end


  # Returns the timestamp for the beginning (00:00:00) of day of the given
  # timestamp.
  def self.beginning_of_day_utc(timestamp)
    timestamp = timestamp.truncate
    timestamp - timestamp % 86400
  end


  # Returns the timestamp for the end (23:59:59) of day of the given
  # timestamp.
  def self.end_of_day_utc(timestamp)
    timestamp = timestamp.truncate
    timestamp - timestamp % 86400 + 86399
  end


  def self.utc_date_string(timestamp)
    t = Time.at(timestamp).gmtime
    sprintf "%s %d, %d UTC", t.strftime("%a %b"), t.mday, t.year
  end


#  WARNING: The sleep() function always returns an integer, but the
#           rounding varies across systems.  If dt is the actual
#           amount of time slept, then sleep() always returns ceil(dt)
#           on some systems, while others always return floor(dt).
#           Although sleep_at_least() isn't meant to be a precision
#           sleep routine, this slight variation in the implementation
#           of sleep() can cause unwanted extra sleep duration (for
#           example, 1.5s longer than requested on a nearly idle
#           2.5GHz CPU), which may be annoying for some uses.
#
#           FreeBSD (4.6 to 7.3), Linux (Xen 2.6.22), and MacOS X
#           (10.5) all seem to precisely sleep with Ruby 1.8.7.
#           The "sleep slop", the extra amount slept beyond requested,
#           is usually around 2 ms on idle or lightly loaded boxes.
#
#  def self.sleep_at_least(sleep_amount)
#    while sleep_amount > 0
#      sleep_amount -= sleep sleep_amount
#    end
#  end

  # Time.now.to_f only takes 2-3 us on a modern CPU.  Even a 400 MHz
  # Pentium II only takes 15 us.
  def self.sleep_at_least(sleep_amount)
    t0 = Time.now.to_f
    while sleep_amount > 0.001   # 1 ms min duration (not worth it if less)
      sleep sleep_amount
      sleep_amount -= Time.now.to_f - t0
    end
  end


  def self.count_with_unit(count, unit, suffix="s")
    retval = "#{count} #{unit}"
    retval += suffix unless count == 1
    retval
  end


  def self.nth(i)
    fail "invalid argument #{i}" if i < 0

    r100 = i % 100
    if r100 >= 10 && r100 < 20
      return "#{i}th"
    else
      r10 = i % 10
      suffix = ["th", "st", "nd", "rd", "th", "th", "th", "th", "th", "th"][r10]
      return i.to_s + suffix
    end
  end


  def self.comma_k(n)
    n == 0 ? "0" : comma(n / 1000) + "k"
  end


  def self.comma(n)
    a = []
    n.to_s.split("").reverse.each_with_index do |d, i|
      a << "," if i > 0 && i % 3 == 0
      a << d
    end
    a.reverse.join("")
  end


  def self.sum(values)
    values.inject(0) { |sum, n| sum += n }
  end


  def self.print_banner(heading)
    puts
    puts "=" * 70
    puts heading
    puts "-" * 70
  end


  # http://en.wikipedia.org/wiki/Fisher-Yates_shuffle
  def self.shuffle_array(array)
    n = array.length
    while n >= 2
      i = rand n
      t = array[i]
      array[i] = array[n - 1]
      array[n - 1] = t
      n -= 1
    end
    array
  end

end  # module ArkUtil
