#############################################################################
## Various utility methods useful for writing interactive command-line
## programs.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: interactiveutils.rb,v 1.4 2008/07/22 19:46:05 youngh Exp $
#############################################################################

module ArkUtil

  # This just prints the {prompt} and a space, so be sure to include any '?'
  # at the end of the prompt yourself.
  #
  # Accepts dates in YYYYMMDD or YYYY-MM-DD format, or timestamps, which
  # must be 9-10 digits long (that is, post Mar 3, 1973).
  def self.prompt_and_get_timestamp(prompt, display_utc=true)
    loop do
      time = nil
      begin
        input = prompt_and_get_line prompt
        if input =~ /^(\d{4})-?(\d{2})-?(\d{2})$/
          time = Time.utc $1, $2, $3, 0, 0, 0, 0
        elsif input =~ /^\d{9,10}$/
          time = Time.at input.to_i
        else
          puts "Invalid input; enter YYYYMMDD, YYYY-MM-DD, or" +
            " 9-10 digit timestamp."
          next
        end
      rescue ArgumentError
        puts "Value is out of range; please enter YYYYMMDD, YYYY-MM-DD," +
          " or 9-10 digit timestamp."
        next
      end

      time.gmtime if display_utc
      puts "You entered " + time.to_s
      input = prompt_and_get_line "Is this correct (yes/no)?", false, true
      return time.to_i if input
    end
  end


  # This just prints the {prompt} and a space, so be sure to include any '?'
  # at the end of the prompt yourself.
  #
  # If {optional} is true, then this will return nil if the user enters
  # a blank line.  Otherwise, this will re-prompt until the user enters
  # a non-blank value.
  #
  # If {yes_no} is true, then this will only allow the input "y"|"yes"|"n"|"no"
  # (regardless of case).  The return result will be true or false.
  def self.prompt_and_get_line(prompt, optional=false, yes_no=false)
    loop do
      print prompt, " "
      $stdout.flush
      input = gets
      unless input
        puts "\nPlease enter some value."
        next
      end

      input.strip!
      if input.length == 0
        if optional
          return input
        else
          puts "Please enter some value."
          next
        end
      end

      if yes_no
        input.downcase!
        if input == "y" || input == "yes"
          return true
        elsif input == "n" || input == "no"
          return false
        else
          puts "Please enter yes/no/y/n."
          next
        end
      end

      return input
    end
  end


  # This accepts a multi-line paragraph of text from the user.
  # {suggested_width} must be >= 2.
  def self.prompt_and_get_paragraph(prompt, suggested_width=0)
    puts prompt
    puts "(Enter '.' on a line by itself to end.)"

    if suggested_width > 0
      leading_text = "(Try to keep lines no longer than "
      trailing_text = ">|)"
      num_dashes = suggested_width - leading_text.length - 2
      if num_dashes > 0
        puts leading_text + ("-" * num_dashes) + trailing_text
      else
        num_dashes = suggested_width - 2
        raise ArgumentError, "suggested_width must be >= 2" if num_dashes < 0
        puts leading_text
        puts ("-" * num_dashes) + trailing_text
      end
    end
    $stdout.flush

    retval = []
    while input = gets
      return retval.join("") if input.chomp == "."
      retval << input
    end

    retval.join("")
  end


  # Executes a block with terminal echo disabled, which is useful for getting
  # passwords from users.
  def self.disable_echo
    require 'termios'

    tio_echo = Termios.tcgetattr($stdin)
    tio_no_echo = tio_echo.dup
    tio_no_echo.lflag &= ~Termios::ECHO
    tio_no_echo.lflag |= Termios::ECHONL

    Termios.tcsetattr($stdin, Termios::TCSANOW, tio_no_echo)
    begin
      yield
    ensure
      Termios.tcsetattr($stdin, Termios::TCSANOW, tio_echo)
    end
  end

end  # module ArkUtil
