#############################################################################
## A hodge-podge of utility methods useful for managing remote processes on
## Ark monitors.
##
## Most of this code is only applicable to very specific situations, but
## the code is needed in several places, so it has been factored out.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008,2009 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: management.rb,v 1.5 2012/03/14 00:23:13 youngh Exp $
#############################################################################

require 'arkutil/monitorutils.rb'
require 'arkutil/sshutils.rb'

module ArkUtil

  def self.ssh_exec!(ssh, command, verbose=false)
    puts ">> " + command if verbose
    output = ssh.exec! command
    puts output if output
  end


  def self.screen_paste(ssh, user, window, text)
    # Without this pausing, screen won't find the .screen.pastebuf.* file for
    # some reason.
    sleep 2

    remote_file = "/home/#{user}/.screen.pastebuf.#{Time.now.to_i}.#{$$}"

    sftp = ssh.sftp
    sftp.file.open(remote_file, "w") do |file|
      file.puts text
    end

    ssh_exec! ssh, "screen -p '#{window}' -X readbuf #{remote_file}"
    ssh_exec! ssh, "screen -p '#{window}' -X paste ."

    sftp.remove! remote_file

    # bug in Net::SFTP causes this to also close the ssh connection
    #sftp.close_channel 
  end


  # Returns the binary-compatibility OS type string for a given OS version.
  def self.os_type(version)
    case version
    when /freebsd7/ then "fbsd71"
    when "freebsd6.2", "freebsd6.4" then "fbsd62"
    when "freebsd4.6" then "fbsd46"
    when /linux/ then "linux"
    else fail "ERROR: unknown OS version " + version
    end
  end


  def self.su_command(user, version)
    case version
    when /freebsd4/ then "sudo su -l " + user
    when /freebsd/, /linux/ then "sudo -i -u " + user
    else fail "ERROR: unknown OS version " + version
    end
  end


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class SudoHelper < ArkUtil::SSHExecHelper

  def initialize
    super(true)  # fail on first ssh exec error
  end

  # Performs sudo su to the ark user inside a screen session on the remote
  # monitor.
  #
  # This is a bit of a hack, since in order to detect the password prompt,
  # we have to connect to screen with a pseudo pty.  Otherwise, we could
  # have just used SudoHelper#exec.
  #
  # NOTE: In order for this hack to detect that the su has succeeded,
  #       the ark user's command prompt MUST start with "(ark@".
  def su_ark_user(ssh, info, window)
    # This is needed so that we don't misinterpret "Password:" strings
    # in the scrollback for pending sudo prompts.
    ssh.exec! "screen -p '#{window}' -X clear"

    sudo_pending = true
    @password_prompt_count = 0
    ssh.open_channel do |channel|
      channel.request_pty do |ch, success|
        abort "ERROR: couldn't obtain pty in su_ark_user()" unless success
        channel.exec("screen -d -r -p '#{window}'") do |ch2, success|
          abort "ERROR: couldn't execute screen -d -r command" unless success

          # NOTE: The sudo password prompt appears on stdout instead of stderr,
          #       unlike when working outside screen.
          channel.on_data do |ch3, data|
            print "OUT>>"
            p data
            channel[:output] ||= ""
            channel[:output] += data
            if channel[:output] =~ /^Password:/ # ^ can match interior line
              channel[:output] = ""
              print "Password:"; $stdout.flush
              channel.send_data password() + "\n"
            elsif channel[:output] =~ /^\(ark@/
              puts "sudo succeeded!"; $stdout.flush
              channel.close
              sudo_pending = false
            elsif channel[:output] =~ /^sudo: \d incorrect password attempt/
              $stderr.puts "sudo failed!"
              exit 1
            end
          end

          channel.on_extended_data do |ch3, type, data|
            print "ERR>>"
            p data
          end
        end

        sudo_command = nil
        case info["os_version"]
        when /freebsd4/ then sudo_command = "sudo su -l ark"
        when /freebsd6/, /freebsd7/, /linux/
          sudo_command = "sudo -i -u ark"
        else abort "INTERNAL ERROR: uknown os_version " + info["os_version"]
        end
        channel.send_data sudo_command + "\n"
      end
    end

    # Note: Just plain ssh.loop doesn't work.  For some reason, it blocks in
    #       certain cases.
    ssh.loop { sudo_pending }
  end

end

end  # module ArkUtil
