#############################################################################
## Various methods useful for working with the Ark monitor spec file
## monitors.yaml.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: monitorutils.rb,v 1.11 2016/10/21 01:19:46 youngh Exp $
#############################################################################

require 'yaml'

require 'arkutil/generalutils'

module ArkUtil

  def self.configure_monitor_spec_option_parser(options, option_parser)
    options.monitor_spec_file =
      if ENV["ARKUTIL_SPEC_PATH"]
        ENV["ARKUTIL_SPEC_PATH"]
      elsif File.exist?("/etc/ark/monitors.yaml")
        "/etc/ark/monitors.yaml"
      else
        "monitors.yaml"
      end

    option_parser.on("--include-monitors", Array, "=LIST",
            "comma-separated monitors to include") do |v|
      options.include_monitors = v
    end

    option_parser.on("--exclude-monitors", Array, "=LIST",
            "comma-separated monitors to exclude") do |v|
      options.exclude_monitors = v
    end

    option_parser.on("--with-ipv6", TrueClass,
                     "match monitors with IPv6") do |v|
      options.with_ipv6 = v
    end

    option_parser.on("--os-version", "=REGEX",
            "regex to match against OS version") do |v|
      options.os_version = v
    end

    option_parser.on("--activity", Array, "=LIST",
                     "comma-separated activities to select") do |v|
      options.activity = v
    end

    option_parser.on("--feature", Array, "=LIST",
                     "comma-separated features to select") do |v|
      options.feature = v
    end

    option_parser.on("--attribute", "=NAME",
                     "name of a monitor attribute to match") do |v|
      options.monitor_attribute = v
    end

    option_parser.on("--value", "=STR",
                     "monitor value to look for") do |v|
      options.monitor_value = v
    end

    option_parser.on("--allow-disabled", TrueClass,
                     "allow selection of disabled monitors") do |v|
      options.allow_disabled = v
    end

    option_parser.on("--list-features", TrueClass,
                     "show available monitor features") do |v|
      options.list_features = v
    end

    option_parser.on("--list-activities", TrueClass,
                     "show available monitor activities") do |v|
      options.list_activities = v
    end

    option_parser.on("--list-monitors", TrueClass,
                     "show available monitors") do |v|
      options.list_monitors = v
    end

    option_parser.on("--spec-file", "=PATH",
            "monitor specification file (#{options.monitor_spec_file})") do |v|
      options.monitor_spec_file = v
    end
  end


  def self.validate_monitor_spec_options(options)
    if options.os_version
      begin
        options.os_version = Regexp.new options.os_version
      rescue
        abort "ERROR: bad value '#{options.os_version}' for --os-version: " +
          $!.to_s
      end
    end

    if !options.monitor_attribute && options.monitor_value
      abort "ERROR: specified --value without --attribute"
    end

    if options.monitor_attribute && !options.monitor_value
      abort "ERROR: specified --attribute without --value"
    end
  end


  def self.process_monitor_spec_options(options)
    all_monitors = load_monitor_spec_file options.monitor_spec_file

    monitors = {}
    all_monitors.each do |monitor, info|
      next unless info["enabled"] || options.allow_disabled

      next if options.include_monitors &&
        !options.include_monitors.include?(monitor)

      next if options.exclude_monitors &&
        options.exclude_monitors.include?(monitor)

      next if options.with_ipv6 && !info["activity_topo_v6"]

      next if options.os_version &&
        !options.os_version.match(info["os_version"])

      if options.activity
        matches = true
        options.activity.each do |activity|
          unless info["activity_" + activity]
            matches = false
            break
          end
        end
        next unless matches
      end

      if options.feature
        matches = true
        options.feature.each do |feature|
          unless info["feature_" + feature]
            matches = false
            break
          end
        end
        next unless matches
      end

      if options.monitor_attribute && options.monitor_value
        value = info[options.monitor_attribute]
        next unless value && value.to_s == options.monitor_value.to_s
      end

      monitors[monitor] = all_monitors[monitor]
    end

    if options.list_features
      values = {}  # feature => true
      monitors.each do |monitor, info|
        next unless info["enabled"] || options.allow_disabled
        info.keys.grep /^feature_(\S+)$/ do |_|
          values[$1] = true
        end
      end

      printf "features: %s\n", values.keys.sort.join(", ")
    end

    if options.list_activities
      values = {}  # activity => true
      monitors.each do |monitor, info|
        next unless info["enabled"] || options.allow_disabled
        info.keys.grep /^activity_(\S+)$/ do |_|
          values[$1] = true
        end
      end

      printf "activities: %s\n", values.keys.sort.join(", ")
    end

    if options.list_monitors
      usable = []
      usable_addrs = []

      monitors.each do |monitor, info|
        next unless info["enabled"] || options.allow_disabled
        usable << [ monitor, info["ip_address"] ]
      end

      usable.sort!
      printf "%d: %s\n", usable.length, usable.map{|x,y| x}.join(",")

      puts
      puts usable.map{|x,y| x + " " + y}

      if options.verbose
        monitors.to_a.sort.each do |monitor, info|
          next unless info["enabled"] || options.allow_disabled
          printf "\n=== %s %s\n", monitor, ("=" * 60)
          info.to_a.sort.each do |k, v|
            printf "  %s: %p\n", k, v
          end
        end
      end
    end

    if options.list_features || options.list_activities || options.list_monitors
      exit 0
    end

    monitors
  end


  def self.load_monitor_spec_file(path)
    defaults = {}  # field => value
    retval = Hash.new { |h,k| h[k] = {} }  # monitor => field => value
    File.open(path) do |file|
      # YAML::each_document no longer exists in Ruby 1.9.3.
      # In Ruby 1.8.7 and 1.9.2, YAML::load_stream returns YAML::Stream (1.8.7)
      # or Syck::Stream (1.9.2), while Ruby 1.9.3p125 returns an Array.
      # The former two have an Array 'documents' attribute with the same
      # data returned by Ruby 1.9.3.
      s = YAML.load_stream file
      s = s.documents if s.respond_to? :documents
      s.each do |document|
        if document[".defaults"]
          defaults = document
          defaults.delete ".defaults"
        elsif document[".monitor"]
          monitor = document.delete ".monitor"
          retval[monitor] = defaults.merge document
        else
          $stderr.puts "ERROR: malformed document in monitor spec file:\n" +
            YAML.dump(document)
          exit 1
        end
      end
    end
    retval
  end

end  # module ArkUtil
