#############################################################################
## A class useful for tracking the set of /24's to probe in team-probing
## Ark measurements.
##
## --------------------------------------------------------------------------
## Copyright (C) 2007 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: probeset.rb,v 1.9 2008/02/20 01:58:35 youngh Exp $
#############################################################################

require 'openssl'
require 'arkutil/iputils'

module ArkUtil

class ProbeSet

  # Set this bit to force OpenSSL::BN to use the maximum required space.
  # Otherwise, calls to {clear_bit!} will fail with OpenSSL::BNError if
  # an attempt is made to clear a bit higher than any set bit.
  KLUDGE_BIT = 2**24

  def initialize(bitmap="")
    @prefix_bitmap = OpenSSL::BN.new bitmap, 2
    @prefix_bitmap.set_bit! KLUDGE_BIT
    @probe_prefixes_gt_24 = Hash.new { |h,k| h[k] = Array.new }
    @noprobe_prefixes_gt_24 = Hash.new { |h,k| h[k] = Array.new }
  end

  # {prefix} should be a string of the form "addr/length".
  def add_prefix(prefix)
    start_addr, end_addr = IPUtils.parse_prefix_to_range prefix
    raise ArgumentError, "invalid prefix '#{prefix}'" unless start_addr

    length = prefix.split("/", 2)[1].to_i
    if length <= 24
      for p in (start_addr >> 8) .. (end_addr >> 8)
	@prefix_bitmap.set_bit! p
      end
    else
      @probe_prefixes_gt_24[start_addr >> 8] <<
	[ start_addr & 0xFF, end_addr & 0xFF ]
    end
  end

  # {prefix} should be a string of the form "addr/length".
  def block_prefix(prefix)
    start_addr, end_addr = IPUtils.parse_prefix_to_range prefix
    raise ArgumentError, "invalid prefix '#{prefix}'" unless start_addr

    length = prefix.split("/", 2)[1].to_i
    if length <= 24
      for p in (start_addr >> 8) .. (end_addr >> 8)
	@prefix_bitmap.clear_bit! p
      end
    else
      @noprobe_prefixes_gt_24[start_addr >> 8] <<
	[ start_addr & 0xFF, end_addr & 0xFF ]
    end
  end

  # {prefix} should be an integer specifying the index of a /24 prefix; that
  # is, {prefix = p >> 8} where p is a /24 prefix.
  def prefix_allowed?(prefix_index)
    @prefix_bitmap.bit_set? prefix_index
  end

  # Returns a randomly-selected non-blocked destination in the given prefix.
  # Registered no-probe prefixes greater than /24 in length are consulted
  # when picking the destination.
  # 
  # The all-zeros and all-ones hosts (that is, .0 and .255) are never returned.
  #
  # {prefix} should be an integer specifying the index of a /24 prefix; that
  # is, {prefix = p >> 8} where p is a /24 prefix.
  def find_random_destination(prefix_index)
    if @noprobe_prefixes_gt_24.has_key? prefix_index
      possible = Array.new.fill(1..254) {|i| i}
      noprobe_list = @noprobe_prefixes_gt_24[prefix_index]
      noprobe_list.each {|x, y| len = y - x + 1; possible.fill(nil, x, len)}
      possible.compact!
      if possible.empty?
	fail "all destinations blocked in prefix index #{prefix_index}"
      end
      (prefix_index << 8) | possible[rand(possible.length)]
    else
      (prefix_index << 8) | (rand(254) + 1)
    end
  end

  # Returns the internal bitmap as raw octets (NOT as ASCII binary string),
  # big-endian order.  This value can be passed to initialize() to recreate
  # the bitmap.
  def bitmap
    @prefix_bitmap.clear_bit! KLUDGE_BIT
    retval = @prefix_bitmap.to_s 2
    @prefix_bitmap.set_bit! KLUDGE_BIT
    retval
  end

end

end  # module ArkUtil
