#############################################################################
## Ark Kafka Producer
##
## --------------------------------------------------------------------------
## Copyright (C) 2024 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
#############################################################################

module ArkUtil

class ArkProducer

  def initialize(server, monitor, topic)
    @server = server
    @monitor = monitor
    @topic = topic
    @cacert = self.class.which_cacert
    @cert = self.class.which_cert(@monitor)
    @key = self.class.which_key(@monitor)
    @seclevel = self.class.which_seclevel(@monitor)

    puts "#{$0} cacert: #{@cacert}"
    puts "#{$0} cert: #{@cert}"
    puts "#{$0} key: #{@key}"

    raise ArgumentError.new("Missing cacert file") if @cacert == nil
    raise ArgumentError.new("Missing cert file") if @cert == nil
    raise ArgumentError.new("Missing key file") if @key == nil

    kafka_config = {
      :"bootstrap.servers" => "#{@server}:#{@seclevel > 0 ? 9095 : 9094}",
      :"client.id" => @monitor,
      :"ssl.key.location" => @key,
      :"ssl.certificate.location" => @cert,
      :"ssl.ca.location" => @cacert,
      :"ssl.cipher.suites" => "DEFAULT@SECLEVEL=#{@seclevel}",
      :"enable.ssl.certificate.verification" => true,
      :"security.protocol" => "ssl",
    }
    @producer = Rdkafka::Config.new(kafka_config).producer
    if block_given?
      yield(self)
      self.close
    end
  end

  def close
    # flush is in the documentation and works on bullseye but not buster?
    # hopefully waiting for each message handle should be enough anyway
    if @producer.respond_to?("flush")
      @producer.flush
    end
    @producer.close
  end

  def produce(data)
    handle = @producer.produce(
      topic: @topic,
      payload: data.to_json,
      key: @monitor,
    )
    # wait until the message is delivered
    handle.wait
  end

  # If attributes of the installed cacert change, then the suffix of the file
  # name change too. Don't want to hardcode all the options, so try using the
  # newest file that matches the glob
  def self.which_cacert
    cacert_paths = [
      "/etc/step/certs/root_ca.crt",
      "/etc/ssl/certs/ca.ark.caida.org_Root_CA_*.pem",
      "/etc/ark/ssl/cacert.pem",
    ]
    Dir.glob(cacert_paths).max_by { |f| File.mtime(f) }
  end

  def self.which_cert(monitor)
    cert_paths = [
      "/etc/ark/ssl/#{monitor}.ark.caida.org.crt",
      "/etc/ark/ssl/client-cert.pem",
    ]
    cert_paths.each do |path|
      if File.exist?(path)
        return path
      end
    end
    return nil
  end

  def self.which_key(monitor)
    key_paths = [
      "/etc/ark/ssl/#{monitor}.ark.caida.org.key",
      "/etc/ark/ssl/client-key.INSECURE.pem",
    ]
    key_paths.each do |path|
      if File.exist?(path)
        return path
      end
    end
    return nil
  end

  def self.which_seclevel(monitor)
    key = which_key(monitor)
    raise ArgumentError.new("Missing key file") if key == nil
    if key.include?("INSECURE")
      return 0
    end
    return 2
  end
end

end
