#############################################################################
## A replacement for Queue in the standard library that uses ArkUtil::List
## instead of an Array, which makes this class more scalable for large numbers
## of elements.
##
## Another benefit to using ArkUtil::List is that Array#shift (which is used
## by Queue) seems to leak memory on certain older Ruby versions.
##
## Note:  Access to this class is synchronized just as with Queue.
##
## --------------------------------------------------------------------------
## Copyright (C) 2009 The Regents of the University of California.
## 
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: queue.rb,v 1.1 2009/05/06 06:56:41 youngh Exp $
#############################################################################

require 'thread'
require 'arkutil/list'

module ArkUtil

class Queue

  def initialize
    @mutex = Mutex.new
    @cond = ConditionVariable.new
    @elements = ArkUtil::List.new
    @num_waiting = 0
  end

  public #==================================================================

  def push(value)
    @mutex.synchronize do
      @elements << value
      @cond.signal if @elements.length == 1 && @num_waiting > 0
    end
  end

  alias << push
  alias enq push

  def pop(non_block=false)
    @mutex.synchronize do
      if @elements.empty?
        raise ThreadError, "queue empty" if non_block
        @num_waiting += 1
        @cond.wait(@mutex) while @elements.empty?
        @num_waiting -= 1
      end
      return @elements.shift
    end
  end

  alias shift pop
  alias deq pop

  def empty?
    @mutex.synchronize do
      return @elements.empty?
    end
  end

  def clear
    @mutex.synchronize do
      return @elements.clear
    end
  end

  def length
    @mutex.synchronize do
      return @elements.length
    end
  end

  alias size length

  def num_waiting
    @mutex.synchronize do
      return @num_waiting
    end
  end

end

end
