#############################################################################
## This class provides a wrapper around rsync that times out and retries
## transfers that stall or fail due to network problems.
##
## --------------------------------------------------------------------------
## Copyright (C) 2007-2008,2012 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: rsync.rb,v 1.12 2014/08/09 00:33:09 youngh Exp $
#############################################################################

require 'logger'
require 'open4'

module ArkUtil

# Rsync behavior notes:
#
#   NOTE: Some versions/configurations of rsync (v3.1.0?) output numbers
#         with commas; e.g., "262,144  81%  248.30kB/s    0:00:00" and
#          "sent 317 bytes  received 5,800,786 bytes  504,443.74 bytes/sec" and
#          "total size is 511,107,783  speedup is 88.11"
#
#         Also, some rsync versions print "receiving incremental file list"
#         instead of "sending incremental file list".
#
#         And print "(xfr#7 to-chk=1/8)" instead of "(xfer#7, to-check=1/8)".
#
# ........................................................................
# ### rsync doesn't print out the file name(s?) if no actual transfer is
#     needed:
# $ rsync -av --progress xyz $SANUS:
# sending incremental file list
#
# sent 76 bytes  received 12 bytes  176.00 bytes/sec
# total size is 5491433  speedup is 62402.65
# ........................................................................
# ### rsync prints progress even when there is an error the sole file being
#     transferred:
# $ rsync -av --bwlimit=50 --progress --timeout=10 nosuchfile $SANUS:
# sending incremental file list
# rsync: link_stat "/Users/youngh/Work/archipelago/util/nosuchfile" failed: No such file or directory (2)
#
# sent 12 bytes  received 12 bytes  16.00 bytes/sec
# total size is 0  speedup is 0.00
# rsync error: some files/attrs were not transferred (see previous errors) (code 23) at main.c(1072) [sender=3.0.8]
# ........................................................................
# ### when no --timeout or --timeout is greater than ssh connection timeout:
#     rsync code 255 is normally a "fail, no retry" but not if the connection
#     was lost:
# ssh: connect to host 196.46.233.22 port 22: Operation timed out
# rsync: connection unexpectedly closed (0 bytes received so far) [sender]
# rsync error: unexplained error (code 255) at io.c(601) [sender=3.0.8]
# ........................................................................
# ### find lib | wc -l => 32 (files & dirs)
# ### after removing 4 files at destination
# $ rsync -av --bwlimit=50 --progress --timeout=10 lib $SANUS:junk/
# sending incremental file list
# lib/arkutil/
# lib/arkutil/scamperdatafeed.rb
#       21824 100%    0.00kB/s    0:00:00 (xfer#1, to-check=9/32)
# lib/arkutil/scamperio.rb
#       10679 100%   65.18kB/s    0:00:00 (xfer#2, to-check=8/32)
# lib/arkutil/sccontrol.rb
#        9555 100%   19.44kB/s    0:00:00 (xfer#3, to-check=7/32)
# lib/arkutil/sshutils.rb
#        9437 100%   15.75kB/s    0:00:00 (xfer#4, to-check=6/32)
# ........................................................................

class RsyncState

  READ_SIZE = 16384        # num bytes to read at once with sysread

  attr_reader :id, :stdout_lines, :stderr,
    :sending_file_list, :progress, :transfer_stats, :total_stats,
    :current_file, :unknown_line

  # options: :debug => bool, :capture_stdout => bool
  def initialize(id, delegate, options={})
    @id = id
    @delegate = delegate
    @options = options
    @debug = options[:debug]
    @slow_attempts = 0
    @fast_attempts = 0
    @stdout_lines = []
    @stderr = nil
    reset_state()
  end

  def start_slow_attempt
    @slow_attempts += 1
    @fast_attempts = 0
    reset_state()
    begin
      @delegate.slow_attempt(self, @slow_attempts) if @delegate
    rescue
      $stderr.printf "RsyncState: @delegate.slow_attempt " +
        "raised exception %s: %s\n%s", $!.class.name, $!.to_s,
        $!.backtrace.join(" <= ")
    end
  end

  def start_fast_attempt
    @fast_attempts += 1
    reset_state()
    begin
      @delegate.fast_attempt(self, @fast_attempts) if @delegate
    rescue
      $stderr.printf "RsyncState: @delegate.fast_attempt " +
        "raised exception %s: %s\n%s", $!.class.name, $!.to_s,
        $!.backtrace.join(" <= ")
    end
  end

  def process_stdout(pout)
    process_output(pout, "stdout") do |original_line|
      $stderr.printf "OUT>> %p\n", original_line if @debug
      @stdout_lines << original_line if @options[:capture_stdout]
      next if original_line == ""

      # Some versions of rsync add commas to numbers, which Ruby doesn't parse
      # correctly; e.g., "1234,234".to_i => 1234; "1234,234".to_f => 1234.0
      line = original_line.delete ","

      case line
      when /incremental file list/
        @sending_file_list = true

      # "       32768   0%    0.00kB/s    0:00:00"
      # "       10679 100%   65.18kB/s    0:00:00 (xfer#2, to-check=8/32)"
      when /^\s+(\d+)\s+(\d+)\%/
        @progress = line
        if @delegate
          bytes, percent = $1.to_i, $2.to_i
          xfer = check_left = check_total = nil
          if line =~ /(?:xfer|xfr)#(\d+),? to-(?:check|chk)=(\d+)\/(\d+)/
            xfer, check_left, check_total = $1.to_i, $2.to_i, $3.to_i
          end

          begin
            @delegate.progress(self, bytes, percent, xfer,
                               check_left, check_total)
          rescue
            $stderr.printf "RsyncState: @delegate.progress raised " +
              "exception %s: %s\n%s", $!.class.name, $!.to_s,
              $!.backtrace.join(" <= ")
          end
        end

      # "sent 5492220 bytes  received 31 bytes  255453.53 bytes/sec"
      when /^sent (\d+) bytes  received (\d+) bytes  (\S+) bytes\/sec$/
        @transfer_stats = line
        begin
          @delegate.transfer_stats(self, $1.to_i, $2.to_i, $3.to_f) if @delegate
        rescue
          $stderr.printf "RsyncState: @delegate.transfer_stats " +
            "raised exception %s: %s\n%s", $!.class.name, $!.to_s,
            $!.backtrace.join(" <= ")
        end

      # "total size is 5491433  speedup is 1.00"
      when /^total size is (\d+)  speedup is (\S+)$/
        @total_stats = line
        begin
          @delegate.total_stats(self, $1.to_i, $2.to_f) if @delegate
        rescue
          $stderr.printf "RsyncState: @delegate.total_stats " +
            "raised exception %s: %s\n%s", $!.class.name, $!.to_s,
            $!.backtrace.join(" <= ")
        end

      else
        if @sending_file_list
          @current_file = original_line
          begin
            @delegate.current_file(self, original_line) if @delegate
          rescue
            $stderr.printf "RsyncState: @delegate.current_file raised " +
              "exception %s: %s\n%s", $!.class.name, $!.to_s,
              $!.backtrace.join(" <= ")
          end
        else
          @unknown_line = original_line
          begin
            @delegate.unknown_line(self, original_line) if @delegate
          rescue
            $stderr.printf "RsyncState: @delegate.unknown_line raised " +
              "exception %s: %s\n%s", $!.class.name, $!.to_s,
              $!.backtrace.join(" <= ")
          end
        end
      end

      dump_state() if @debug
    end
  end

  def process_stderr(perr)
    @stderr = perr.read
    $stderr.printf "ERR>> %p\n", @stderr if @debug
  end


  private # ..............................................................

  def reset_state
    @sending_file_list = false
    @progress = nil
    @transfer_stats = nil
    @total_stats = nil
    @current_file = nil
    @unknown_line = nil
  end


  def dump_state
    $stderr.printf "@sending=%p; @prog=%p; @trans=%p; @total=%p; @cur=%p; @unkw=%p\n",
      @sending_file_list, @progress, @transfer_stats, @total_stats,
      @current_file, @unknown_line
  end


  def process_output(io, io_name, &b)
    line_buffer = []  # [ String ]: for accumulating line fragments
    data = ""
    begin
      loop do
        io.sysread READ_SIZE, data

        # Note: String#split treats trailing \r or \n differently than
        #       embedded \r|\n, so that case must be handled specially:
        #         >> "foo\n".split(/\r|\n/) => ["foo"]
        #         >> "foo\r\n".split(/\r|\n/) => ["foo"]
        lines = data.split /\r|\n/
        if @debug
          $stdout.printf "process_%s: %p => %p\n", io_name, data, lines
        end

        line_buffer << lines.first
        if lines.length > 1
          b.call line_buffer.join("")
          line_buffer.clear

          (1 .. lines.length - 2).each do |i|
            b.call lines[i]
          end

          line_buffer << lines.last
        end

        # Handle the special case of data ending in \r|\n.
        c = data[-1, 1]
        if c == "\n" || c == "\r"
          b.call line_buffer.join("")
          line_buffer.clear
        end        
      end

    rescue EOFError
      unless line_buffer.empty?
        b.call line_buffer.join("")
        line_buffer.clear
      end

    rescue IOError, SystemCallError
      if @debug
        $stdout.printf "process_%s: %s: %s\n", io_name,
        $!.class.name, $!.inspect
      end
      
      # XXX handle this

    end
  end

end


#============================================================================

class Rsync

  class StderrPrinter
    def method_missing(name, *args)
      args[0] = name.to_s.upcase! + ": " + args[0] + "\n"
      $stderr.printf *args
    end
  end

  # Data transfer is subject to two levels of timeouts and retries, an outer
  # 'slow' retry and an inner 'fast' retry, as follows:
  #
  #   -------------------------------------------------------------------
  #   * attempt 1
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #   * 'slow' retry: attempt 2, retrying after 15 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #   * 'slow' retry: attempt 3, retrying after 60 minutes
  #      ...
  #
  #   * 'slow' retry: attempt 4, retrying after 4 hours
  #      ...
  #
  #   * 'slow' retry: attempt 5, 6, ..., N, retrying after 16 hours
  #      ...
  #   -------------------------------------------------------------------
  #
  # The 'slow' retry attempts are spread relatively far apart and are meant
  # to be on the timescale of equipment failure-fix cycles.  The 'fast'
  # retry attempts are meant to be on the timescale of transient network
  # problems (e.g., temporary routing loops, congestion-based loss/collapse,
  # route loss during BGP path exploration).
  #
  # Retries happen for two reasons, failure to open a connection and failure
  # to make progress on a transfer.  Currently, "failure to make progress" is
  # defined simplisitically as any period of N consecutive seconds with 0
  # bytes transferred.  A more sophisticated definition might be based on
  # a lower bound on the transfer rate as computed with an exponential moving
  # average.
  #
  # NOTE: If a transfer fails to make progress, the retry attempt happens
  #       in the outer 'slow' retry loop rather than in the inner 'fast'
  #       retry loop.  The purpose of the inner 'fast' retry loop is to
  #       react quickly to connection loss and failure to open a connection.
  #       If a transfer fails to make progress, then we typically want to
  #       react more slowly (that is, retry after a moderately long delay),
  #       so the inner 'fast' retry loop is aborted, and the transfer is
  #       retried in the outer 'slow' loop.  (An important point is that
  #       failure to make progress is detected at all times, regardless of
  #       at what stage the transfer is at.)
  # 
  # Failure to open a connection ("open-conn failure") is handled separately
  # from the other failures that happen once a connection is opened and the
  # transfer started.  An open-conn failure is retried at a different
  # configured retry rate than either the 'slow' or 'fast' retry rates.
  # Furthermore, an open-conn failure is retried indefinitely, whereas the
  # 'slow' and 'fast' retries are limited to a finite number of attempts.
  # Retries for open-conn failures aren't counted toward the 'slow' retries
  # even though retries for open-conn failures happen at the same level as
  # 'slow' retries, as shown below:
  #
  #   -------------------------------------------------------------------
  #   * attempt 1
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #      [connection couldn't be opened even after 3 'fast' retries]
  #
  #   * retrying open connection after 15 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #      [connection couldn't be opened even after 3 'fast' retries]
  #
  #   * retrying open connection after 60 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #      [connection couldn't be opened even after 3 'fast' retries]
  #
  #   * retrying open connection after 4 hours
  #      ...
  #
  #      [some failure besides failure to open connection]
  #
  #   * 'slow' retry: attempt 2, retrying after 15 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #   -------------------------------------------------------------------
  #
  # NOTE: 'fast' and 'slow' retries are based on different

  #=========================================================================

  # Parameters for 'slow' retries.
  #
  # The slow timeout starts at INITIAL_SLOW_TIMEOUT and increases by a
  # factor of SLOW_TIMEOUT_MULTIPLIER at each attempt, capping at
  # MAX_SLOW_TIMEOUT.  No more than MAX_SLOW_ATTEMPTS attempts will
  # be made.
  #
  # Although not strictly required, it makes the most sense for the
  # parameters for 'slow' retries to be larger than those for 'fast' retries
  # (given below).  In particular, INITIAL_SLOW_TIMEOUT should be greater
  # than the total amount of delay possible with 'fast' retries after the
  # maximum allowed number of retries.  If INITIAL_SLOW_TIMEOUT is set too
  # low, then it is possible for the 'slow' retry loop to interrupt the
  # 'fast' retry loop.  This could be the desired behavior in some cases,
  # but in most cases, it probably will not be.

  @@default_options = {}

  @@default_options[:initial_slow_timeout] = 15 * 60 # seconds
  @@default_options[:max_slow_timeout] = 2 * 3600 # seconds
  @@default_options[:slow_timeout_multiplier] = 2
  @@default_options[:max_slow_attempts] = 15  # ~24hr max attempt duration

  # suggested values for testing purposes
  # INITIAL_SLOW_TIMEOUT = 60 # seconds
  # MAX_SLOW_TIMEOUT = 120 # seconds
  # SLOW_TIMEOUT_MULTIPLIER = 2
  # MAX_SLOW_ATTEMPTS = 3
  # WATCHDOG_TIMEOUT = 10 # seconds

  # Parameters for handling failures to open connections (open-conn failures).
  #
  # Unlike 'slow' and 'fast' retries, open-conn failures are retried
  # indefinitely.  The delay between retries starts at INITIAL_CONNECT_DELAY
  # and increases by a factor of CONNECT_DELAY_MULTIPLIER, capping at
  # MAX_CONNECT_DELAY.

  @@default_options[:initial_connect_delay] = 120 # seconds
  @@default_options[:max_connect_delay] = 240 # seconds
  @@default_options[:connect_delay_multiplier] = 2

  # suggested values for testing purposes
  # INITIAL_CONNECT_DELAY = 15 # seconds
  # MAX_CONNECT_DELAY = 30 # seconds
  # CONNECT_DELAY_MULTIPLIER = 2

  # Parameters for 'fast' retries.
  #
  # The delay between retries starts at INITIAL_FAST_DELAY and
  # increases by a factor of FAST_DELAY_MULTIPLIER at each attempt,
  # capping at MAX_FAST_DELAY.  No more than MAX_FAST_ATTEMPTS
  # attempts will be made.
  #
  # Generally speaking, these parameter values should be lower than the
  # values for 'fast' retries.  See the comments at the latter for details.

  @@default_options[:initial_fast_delay] = 4 # seconds
  @@default_options[:max_fast_delay] = 16 # seconds
  @@default_options[:fast_delay_multiplier] = 4
  @@default_options[:max_fast_attempts] = 3


  #=========================================================================

  # Exit codes:
  #
  #   0 -- everything was successful; the only possible non-success is a
  #        warning that the file couldn't be removed from the server
  #
  #   1 -- file doesn't exist on the server
  #
  #   2 -- file downloaded successfully but some error happened afterwards
  #        (e.g., couldn't rename <file>.download to <file>); don't retry
  #        download, and generally prefer to fail hard and defer to human
  #        intervention
  #
  #   3 -- file not downloaded successfully and some non-fatal error happened;
  #        may retry
  #
  #   4 -- file not downloaded successfully and some fatal error happened;
  #        don't retry
  EXIT_SUCCESS = 0
  EXIT_NO_FILE = 1
  EXIT_FAILED_BUT_RETRY = 2
  EXIT_FAILED_DONT_RETRY = 3
  EXIT_COULDNT_CONNECT_BUT_RETRY = 4

  #=========================================================================

  def initialize(options={}, syslog=nil)
    @options = @@default_options.dup
    @options.merge! options
    @debug = @options[:debug]

    @log = syslog || StderrPrinter.new
  end


  def execute(command, state)
    execute_rsync_slow_attempt(command, state)
  end


  private #-----------------------------------------------------------------

  # Returns * true on success,
  #         * false if the remote file doesn't exist, and
  #         * {error string} if transfer fatally failed, possibly after
  #           multiple retries.
  #
  # This method prefers to exit on certain errors rather than trying to
  # recover, since it's best for a human to intervene in some cases, even
  # when the error wouldn't necessarily be considered fatal by everyone.
  # This "fail-fast" approach does, however, depend on having multiple
  # collection processes running concurrently, so that data collection
  # doesn't completely halt.

  def execute_rsync_slow_attempt(command, state)
    connect_delay = @options[:initial_connect_delay]
    timeout = @options[:initial_slow_timeout]

    current_try = 1
    loop do
      state.start_slow_attempt()
      result = execute_rsync_fast_attempt command, state

      case result
      when EXIT_SUCCESS
        return true

      when EXIT_NO_FILE
        return false

      when EXIT_FAILED_BUT_RETRY
        @log.info("Rsync:slow_attempt: rsync failed " +
                  "but we can retry; command=%p", command)

        if current_try == @options[:max_slow_attempts]
          msg = sprintf("Rsync:slow_attempt: reached maximum number of " +
                        "'slow' attempts; aborting; command=%p", command)
          @log.err("%s", msg)
          return msg
        else
          current_try += 1
          @log.info("Rsync:slow_attempt: doing 'slow' retry " +
                    "attempt %d after %d seconds; command=%p", 
                    current_try, timeout, command)

          sleep_at_least timeout
          timeout = [@options[:max_slow_timeout],
                     timeout * @options[:slow_timeout_multiplier]].min
        end

      when EXIT_FAILED_DONT_RETRY
        msg = sprintf("Rsync:slow_attempt: rsync had a fatal error; " +
                      "aborting; command=%p", command)
        @log.err("%s", msg)
        return msg

      when EXIT_COULDNT_CONNECT_BUT_RETRY  # XXX
        @log.info("Rsync:slow_attempt: couldn't connect to '%s'; " +
                  "retrying after %d seconds", @full_host, connect_delay)

        sleep_at_least connect_delay
        connect_delay = [@options[:max_connect_delay],
                       connect_delay * @options[:connect_delay_multiplier]].min

      else
        msg = sprintf("Rsync:slow_attempt: INTERNAL ERROR: invalid " +
                      "result %p; aborting; command=%p", result, command)
        @log.err("%s", msg)
        return msg
      end
    end
    # NOTREACHED
  end


  # Rsync return codes:
  #
  # ........................................................................
  # ### when no --timeout or --timeout is greater than ssh connection timeout:
  # ssh: connect to host 196.46.233.22 port 22: Operation timed out
  # rsync: connection unexpectedly closed (0 bytes received so far) [sender]
  # rsync error: unexplained error (code 255) at io.c(601) [sender=3.0.8]
  # ........................................................................
  # ### when --timeout is less than the ssh connection timeout:
  # $ time rsync -av --progress --timeout=10 scamper-cvs-20110803-fbsd46.tar.gz $BJLGM:
  # [sender] io timeout after 10 seconds -- exiting
  # rsync error: timeout in data send/receive (code 30) at io.c(140) [sender=3.0.8]
  # ........................................................................
  # ### local chdir failure:
  # rsync: change_dir "/Users/youngh/Work/programming/rsync/rsync-3.0.9/../team-probing" failed: No such file or directory (2)
  #
  # sent 12 bytes  received 12 bytes  48.00 bytes/sec
  # total size is 0  speedup is 0.00
  # rsync error: some files/attrs were not transferred (see previous errors) (code 23) at main.c(1072) [sender=3.0.8]
  # ........................................................................
  # ^CKilled by signal 2.
  # rsync error: unexplained error (code 255) at rsync.c(657) [sender=3.0.8]
  # ........................................................................
  # ### hitting ^C at password prompt (which can happen for invalid username):
  # Password:^C
  # rsync error: received SIGINT, SIGTERM, or SIGHUP (code 20) at rsync.c(657) [sender=3.0.8]
  # ........................................................................
  # rsync: connection unexpectedly closed (9 bytes received so far) [sender]
  # rsync error: error in rsync protocol data stream (code 12) at io.c(601) [sender=3.0.7]
  # ........................................................................
  # rsync: mkdir "/public-data/........." failed: Permission denied (13)
  # rsync error: error in file IO (code 11) at main.c(577) [Receiver=3.0.7]
  # rsync: connection unexpectedly closed (9 bytes received so far) [sender]
  # rsync error: error in rsync protocol data stream (code 12) at io.c(601) [sender=3.0.7]
  # ........................................................................


  # $ cat ~/Work/programming/rsync/rsync-3.0.9/errcode.h | perl -ne 'if (/^\#define (\S+)\s+(\d+)(.*)$/) { my ($code,$num,$rest)=($1,$2,$3); $rest =~ s:/\*:#:; $rest =~ s: \*/::; print "$code=$num$rest\n" } else { $_ =~ s:^/\*| \*(/)?:\# :; print }'

  RERR_OK=0
  RERR_SYNTAX=1       # syntax or usage error
  RERR_PROTOCOL=2       # protocol incompatibility
  RERR_FILESELECT=3       # errors selecting input/output files, dirs
  RERR_UNSUPPORTED=4      # requested action not supported
  RERR_STARTCLIENT=5      # error starting client-server protocol

  RERR_SOCKETIO=10      # error in socket IO
  RERR_FILEIO=11      # error in file IO
  RERR_STREAMIO=12      # error in rsync protocol data stream
  RERR_MESSAGEIO=13      # errors with program diagnostics
  RERR_IPC=14      # error in IPC code
  RERR_CRASHED=15      # sibling crashed
  RERR_TERMINATED=16      # sibling terminated abnormally

  RERR_SIGNAL1=19      # status returned when sent SIGUSR1
  RERR_SIGNAL=20      # status returned when sent SIGINT, SIGTERM, SIGHUP
  RERR_WAITCHILD=21      # some error returned by waitpid()
  RERR_MALLOC=22      # error allocating core memory buffers
  RERR_PARTIAL=23      # partial transfer
  RERR_VANISHED=24      # file(s) vanished on sender side
  RERR_DEL_LIMIT=25      # skipped some deletes due to --max-delete

  RERR_TIMEOUT=30      # timeout in data send/receive
  RERR_CONTIMEOUT=35      # timeout waiting for daemon connection

  #  Although it doesn't seem to be specified anywhere,
  #  ssh and the shell seem to return these values:
  # 
  #  124 if the command exited with status 255
  #  125 if the command is killed by a signal
  #  126 if the command cannot be run
  #  127 if the command is not found
  # 
  #  and we could use this to give a better explanation if the remote
  #  command is not found.
  # 
  RERR_CMD_FAILED=124
  RERR_CMD_KILLED=125
  RERR_CMD_RUN=126
  RERR_CMD_NOTFOUND=127

  #--------------------------------------------------------------------------

  RSYNC_ERROR_HANDLING = {}  # RERR_* to EXIT_*

  RSYNC_ERROR_HANDLING[RERR_OK] = EXIT_SUCCESS

  # ........................................................................

  # RERR_VANISHED: XXX supposedly not an error, but it doesn't seem good to
  # silently treat this case as a success, since it may surprise the user.
  # log.c:log_exit(): /* VANISHED is not an error, only a warning */

  RSYNC_ERROR_HANDLING[RERR_VANISHED] = EXIT_NO_FILE

  # ........................................................................

  # XXX RERR_FILESELECT seems to be for unrecoverable errors (e.g., invalid
  # source/destination paths)

  # XXX disconnection might trigger RERR_STARTCLIENT?

  # RERR_MESSAGEIO seems to be used for various internal programming errors.

  # RERR_PARTIAL => some files/attrs were not transferred (see previous errors)
  # RERR_PARTIAL seems to be a complex error, so let the user figure out how
  # best to proceed.

  # XXX RERR_CMD_RUN and RERR_CMD_NOTFOUND suggest some sort of configuration
  # problem, so getting the user involved seems the best option.

  [ RERR_SYNTAX, RERR_PROTOCOL, RERR_FILESELECT, RERR_UNSUPPORTED,
    RERR_STARTCLIENT, RERR_MESSAGEIO, RERR_PARTIAL, RERR_DEL_LIMIT,
    RERR_CMD_RUN, RERR_CMD_NOTFOUND ].each do |code|
    RSYNC_ERROR_HANDLING[code] = EXIT_FAILED_DONT_RETRY
  end 

  # ........................................................................

  # RERR_IPC seems to indicate a failure of fork, close, pipe, etc., which
  # may happen from transient resource exhaustion.

  # RERR_SIGNAL1 and RERR_SIGNAL indicate signals to the local rsync process
  # from the user.  XXX not sure that proceeding is the best option.

  # XXX RERR_CMD_FAILED and RERR_CMD_KILLED suggest a buggy ssh/shell command;
  # so retry.

  # fail, but retry
  [ RERR_SOCKETIO, RERR_FILEIO, RERR_STREAMIO, RERR_IPC, RERR_CRASHED,
    RERR_TERMINATED, RERR_SIGNAL1, RERR_SIGNAL, RERR_WAITCHILD,
    RERR_MALLOC, RERR_TIMEOUT, RERR_CONTIMEOUT, RERR_CMD_FAILED,
    RERR_CMD_KILLED ].each do |code|
    RSYNC_ERROR_HANDLING[code] = EXIT_FAILED_BUT_RETRY
  end


  #--------------------------------------------------------------------------

  def execute_rsync_fast_attempt(command, state)
    delay = @options[:initial_fast_delay]
    current_try = 1

    loop do
      begin
        state.start_fast_attempt()
        raw_status = Open4.popen4ext(true, command) do |pid, pin, pout, perr|
          threads = []
          threads << Thread.new(pout) do |pout|
            state.process_stdout pout
          end

          threads << Thread.new(perr) do |perr|
            state.process_stderr perr
          end

          threads.each do |thr|
            $stderr.printf "joining stdout/stderr threads\n" if @debug
            thr.join
          end
        end
        p raw_status if @debug
        status = extract_exit_status raw_status
        p status if @debug

      rescue StandardError => exn
        raise exn if @debug  # XXX
        status = [ :exception, exn.inspect ]
      end

      case status[0]
      when :exited
        code = status[1]

        # ssh: connect to host 196.46.233.22 port 22: Operation timed out
        # rsync: connection unexpectedly closed ...
        # rsync error: unexplained error (code 255) at io.c(601) [sender=3.0.8]
        if code == 255
          err = state.stderr || ""
          if err =~ /connect to host|connection unexpectedly closed/
            status[1] = code = RERR_TIMEOUT
          end
        end

        result = RSYNC_ERROR_HANDLING[code]
        case result
        when EXIT_SUCCESS, EXIT_NO_FILE, EXIT_FAILED_DONT_RETRY
          return result

        when EXIT_FAILED_BUT_RETRY, EXIT_COULDNT_CONNECT_BUT_RETRY
          if current_try == @options[:max_fast_attempts]
            @log.info("Rsync:fast_attempt: reached maximum number of " +
                      "'fast' attempts; aborting; command=%p", command)
            return result
          else
            current_try += 1
            @log.info("Rsync:fast_attempt: doing 'fast' retry attempt %d " +
                      "after %d seconds; command=%p", current_try, delay,
                      command)

            sleep_at_least delay
            delay = [@options[:max_fast_delay],
                     delay * @options[:fast_delay_multiplier]].min
          end

        else
          # XXX report unknown rsync code
          return EXIT_FAILED_DONT_RETRY
        end

      when :exception then return EXIT_FAILED_DONT_RETRY  # XXX
      when :signaled then return EXIT_FAILED_DONT_RETRY  # XXX
      when :stopped then return EXIT_FAILED_DONT_RETRY  # XXX shouldn't happen
      else
        return EXIT_FAILED_DONT_RETRY  # XXX
      end

      if false
        @log.err("Rsync:fast_attempt: couldn't connect to '%s': %s: %s",
                 @full_host, $!.class.name, $!.to_s)
        exit! EXIT_FAILED_DONT_RETRY

        @log.err("Rsync:fast_attempt: couldn't connect to '%s': " +
                 "%s: %s", @full_host, $!.class.name, $!.to_s)
        exit_code = EXIT_COULDNT_CONNECT_BUT_RETRY

        @log.err("Rsync:fast_attempt: lost connection to '%s': %s: %s",
                 @host, $!.class.name, $!.to_s)
        exit_code = EXIT_FAILED_BUT_RETRY

        @log.err("Rsync:fast_attempt: got exception %s: %s; " +
                 "backtrace: %s", @full_host, remote_file, $!.class.name,
                 $!.to_s, $!.backtrace.join(" <= "))
        exit_code = EXIT_FAILED_BUT_RETRY
      end

    end
  end


  def extract_exit_status(status)
    retval = []
    if status.exited?
      retval.push :exited, status.exitstatus
    elsif status.signaled?
      retval.push :signaled, status.termsig
    elsif status.stopped?
      retval.push :stopped, status.stopsig
      # elsif status.coredump?
      # Don't return info about status.coredump?, since that only provides
      # additional information for the status.signaled? case.
    end
    retval
  end


  def sleep_at_least(sleep_amount)
    while sleep_amount > 0
      sleep_amount -= sleep sleep_amount
    end
  end

end

end  # module ArkUtil
