#############################################################################
## A class for asynchronously interacting with Scamper over its control
## socket, especially to read warts data over the control socket.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008,2009,2010 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: scamperio.rb,v 1.15 2020/03/26 06:20:59 youngh Exp $
#############################################################################

require 'socket'

# Compatibility for Ruby versions before 1.9.2 preview 2.
class IO
  module WaitReadable; end
  module WaitWritable; end
end

module ArkUtil

class ScamperIO

  class StderrPrinter
    def method_missing(name, *args)
      args[0] = name.to_s.upcase! + ": " + args[0] + "\n"
      $stderr.printf *args
    end
  end

  attr_accessor :debug_io_bytes, :debug_commands
  attr_accessor :client

  READ_SIZE = 16384   # num bytes to read at once with sysread

  # The client should set the {client} attribute before starting up ScamperIO.
  # The client receives several notifications (via callbacks):
  #
  #  * process_scamper_response(response)
  #      where {response} is a newline-terminated multi-line string:
  #
  #         OK
  #         ERR ...
  #         MORE
  #         DATA \d+\n....
  #
  #  * lost_scamper_connection()
  #       when scamper process dies, some fatal error occurs during
  #       communication with scamper, or the connection is dropped as
  #       a result of a shutdown request by the client (via "DONE" command)
  #
  #  * scamperio_failed(message)
  #       when some unknown error happens.
  #
  # The {format} argument should be "warts" or "json", and it's only
  # supported by CAIDA's custom scamper build, for the moment (2017-06-09).
  # This specifies the format of the data returned by scamper over the
  # control socket.
  #
  # The {syslog} argument should be an instance of Syslog.
  def initialize(port, format=nil, syslog=nil)
    @client = nil # allowed to be nil, so long as @client is set eventually
    @port = port
    @log = syslog || StderrPrinter.new

    @sock = TCPSocket.new '127.0.0.1', @port  # propagate exceptions out
    @sock.setsockopt Socket::IPPROTO_TCP, Socket::TCP_NODELAY, true

    @read_buffer = []  # [String]
    @write_queue = []  # [String]
    @response_buffer = nil   # accumulates scamper DATA response lines

    @awaiting_attach_response = true
    attach_command = "attach"
    attach_command += sprintf(" format %s", format) if format
    attach_command += "\n"
    @write_queue << attach_command
  end


  def write_command(text)
    unless @sock
      @log.debug "CONNECTION CLOSED: write_command: text=%p", text
      return
    end

    if @debug_io_bytes
      @log.debug "write_command: text=%p", text
    end

    @write_queue << text + "\n"
  end


  # Methods called by the event loop.

  def io
    @sock
  end


  def want_read
    @sock != nil
  end


  def want_write
    !@write_queue.empty?
  end


  def read_data
    begin
      data = @sock.read_nonblock READ_SIZE
      @log.debug "read_data from %p: %p", @sock, data if @debug_io_bytes

      process_input data

    # Ruby 1.9.2 preview 2 uses IO::WaitReadable, but earlier versions use
    # plain Errno::EWOULDBLOCK, so technically we could get rid of
    # IO::WaitReadable.
    rescue Errno::EINTR, Errno::EAGAIN, Errno::EWOULDBLOCK, IO::WaitReadable
      # do nothing, since we'll automatically retry in the next select() round

    rescue SocketError, IOError, EOFError, SystemCallError
      @log.info "read_data from %p: %p", @sock, $!
      reset_connection()
    end
  end


  def write_data
    if @write_queue.length == 1
      buffer = @write_queue.first
    elsif @write_queue.length > 1
      buffer = @write_queue.join nil
      @write_queue.clear
      @write_queue << buffer
    else  # @write_queue.length == 0 
      return  # nothing to do -- spurious write readiness
    end

    begin
      while buffer.length > 0
        n = @sock.write_nonblock buffer
        data_written = buffer.slice! 0, n
        if @debug_io_bytes
          @log.debug "write_data to %p: wrote %d bytes, %d left: %p",
            @sock, n, buffer.length, data_written
        end
      end

      @write_queue.shift

    # Ruby 1.9.2 preview 2 uses IO::WaitWritable, but earlier versions use
    # plain Errno::EWOULDBLOCK, so technically we could get rid of
    # IO::WaitWritable.  However, we need IO::WaitWritable for SSL.
    rescue Errno::EINTR, Errno::EAGAIN, Errno::EWOULDBLOCK, IO::WaitWritable
      # do nothing, since we'll automatically retry in the next select() round

    # syswrite may throw "not opened for writing (IOError)"
    rescue SocketError, IOError, SystemCallError # including Errno::EPIPE
      @log.info "write_data to %p: %p", @sock, $!
      reset_connection()
    end
  end


  private #=================================================================

  def reset_connection
    @sock.close rescue nil
    @sock = nil
    @read_buffer.clear
    @write_queue.clear
    @response_buffer = nil

    @client.lost_scamper_connection()
  end


  # @read_buffer accumulates partial lines until we read one or more
  # newlines.  Once we read a newline, we split @read_buffer into lines (in
  # case we get a whole bunch of short lines from a single read) and invoke
  # process_scamper_response_line() on each line.  This simplifies the
  # implementation of process_scamper_response_line().
  #
  # Note: This method may discard blank lines (caused by adjacent newlines),
  #       but this is fine for scamper, since scamper never uses blank lines
  #       as delimiters or as data.  This discarding is caused by String#split,
  #       which discards any number of trailing newlines but not consecutive
  #       newlines when they occur at either the beginning or middle of
  #       a string.  The way @read_buffer is managed means a run of newlines
  #       may or may not get dicarded depending on a number of factors.
  def process_input(data)
    # Note: String#split is problematic when there is a trailing newline
    #       ("foo\n".split("\n") => ["foo"]), so we handle that case later.
    lines = data.split "\n"
    if @debug_commands; then p data; p lines; end

    @read_buffer << lines.first
    if lines.length > 1
      process_scamper_response_line @read_buffer.join("")
      @read_buffer.clear

      (1 .. lines.length - 2).each do |i|
        process_scamper_response_line lines[i]
      end

      @read_buffer << lines.last
    end

    if data[-1, 1] == "\n"  # then ends with newline
      process_scamper_response_line @read_buffer.join("")
      @read_buffer.clear
    end        
  end


  def process_scamper_response_line(line)
    case line
    when /^\s*$/ then return
    when "`"
      if @response_buffer
        @client.process_scamper_response @response_buffer + line + "\n"
      else
        @log.err "ScamperIO: received closing line of DATA block when no " +
          "DATA blocks are active"
        reset_connection()
        return
      end
      @response_buffer = nil

    when /^DATA \d+$/
      if @response_buffer
        @log.err "ScamperIO: received starting line of a new DATA block " +
          "even though the active DATA block hasn't ended"
        reset_connection()
        return
      end
      @response_buffer = line + "\n"

    else
      if @response_buffer  # currently in DATA block
        @response_buffer += line + "\n"

        # Note: The only legal uuencode characters are space (32) to underline
        #       (92); specifically, opening curly is not a legal character.
        if line.start_with? "{\""  # start of JSON output
          @client.process_scamper_response @response_buffer
          @response_buffer = nil # JSON output only has a single line after DATA
        end
      elsif @awaiting_attach_response
        @awaiting_attach_response = false
        if line == "ERR"
          reset_connection()
        end # else silently gobble up "OK"
      else
        @client.process_scamper_response line + "\n"
      end
    end
  end

end

end  # module ArkUtil

