#############################################################################
## A class for interacting with scamper through its control socket.
##
## --------------------------------------------------------------------------
## Copyright (C) 2007 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: sccontrol.rb,v 1.25 2013/05/03 00:57:51 youngh Exp $
#############################################################################

require 'socket'

module ArkUtil

class ScamperControl

  class ConnectionBroken < IOError; end

  private

  def self.scamper_attr_reader(*symbols)
    symbols.each do |symbol|
      eval <<-EOF
      def #{symbol.id2name}
	get "#{symbol.id2name}"
      end
      EOF
    end
  end

  def self.scamper_int_attr_reader(*symbols)
    symbols.each do |symbol|
      eval <<-EOF
      def #{symbol.id2name}
	value = get "#{symbol.id2name}"
	value && value.to_i
      end
      EOF
    end
  end

  def self.scamper_attr_writer(*symbols)
    symbols.each do |symbol|
      eval <<-EOF
      def #{symbol.id2name}=(value)
	set "#{symbol.id2name}", value
      end
      EOF
    end
  end

  #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  def initialize(port)
    @sock = TCPSocket.open '127.0.0.1', port
    @sock.setsockopt Socket::IPPROTO_TCP, Socket::TCP_NODELAY, true
  end

  # Raises ConnectionBroken if send fails with EPIPE.
  # XXX need to investigate what other errors can happen when connection is
  #     lost and pass these errors out as ConnectionBroken
  def sock_send(command)
    $stderr.puts "sock_send: #{command}"
    begin
      length = command.length
      while length > 0
	sent = @sock.send command, 0
	length -= sent
	command = command[sent .. -1] if length > 0
      end
    rescue # Errno::EPIPE
      raise_connection "lost connection: #{$!.class.name}: #{$!}"
    end
  end

  # Returns an array of response lines without \n.
  #
  # NOTE: This code assumes that only *one* response is unread from scamper.
  #       This assumption exists in the detection of the terminating "OK"
  #       or "ERR".  You should not issue multiple commands with sock_send
  #       and then call sock_read_response to read back the responses one
  #       at a time because sock_read_response will return just one
  #       (malformed) response, not several individual ones.
  #
  # Raises ConnectionBroken on (premature) EOF.
  def sock_read_response
    begin
      retval = []
      loop do
	line = @sock.gets
        raise_connection "read EOF" unless line

	retval << line.chomp
	return retval if line =~ /^(OK|ERR)/
      end
    rescue ConnectionBroken
      raise
    rescue # Errno::ECONNRESET
      raise_connection "lost connection: #{$!.class.name}: #{$!}"
    end
  end

  def do_command(command)
    sock_send command
    response = sock_read_response
    response[0] =~ /^OK/ ? true : false
  end

  def raise_connection(message)
    @sock.close rescue nil if @sock
    @sock = nil
    raise ConnectionBroken, message
  end

  public #===================================================================

  def get(attribute)
    sock_send "get #{attribute}\n"
    response = sock_read_response
    response[0] =~ /^OK \S+ (.+)$/
    $1
  end

  scamper_int_attr_reader :attempts, :dport, :holdtime, :hoplimit, :pps, :pid
  scamper_int_attr_reader :sport, :timeout
  scamper_attr_reader :version

  # Returns the newly set value on success or nil on error.
  def set(attribute, value)
    sock_send "set #{attribute} #{value}\n"
    response = sock_read_response
    response[0] =~ /^OK \S+ (.+)$/
    $1
  end

  scamper_attr_writer :attempts, :holdtime, :hoplimit, :pps, :timeout

  def open_outfile(name, path, mode="append")
    do_command "outfile open name \"#{name}\" file \"#{path}\" mode #{mode}\n"
  end

  def close_outfile(name)
    do_command "outfile close \"#{name}\"\n"
  end

  # Returns list of [ <name>, <path>, <refcount> ] or nil.
  def list_outfiles
    sock_send "outfile list\n"
    response = sock_read_response
    return nil if response[-1] =~ /^ERR/

    retval = []
    response.each_line do |line|
      next unless line =~ /^INFO '(.+)' file '(.+)' refcnt (\d+)$/
      retval << [ $1, $2, $3 ]
    end
    retval
  end

  # Swaps the file paths associated with {name1} and {name2}; e.g.,
  #
  #   > outfile list
  #   INFO 'out1' file 'tmp_out1.warts' refcnt 0
  #   INFO 'out2' file 'tmp_out2.warts' refcnt 0
  #   > outfile swap out1 out2
  #   > outfile list
  #   INFO 'out1' file 'tmp_out2.warts' refcnt 0
  #   INFO 'out2' file 'tmp_out1.warts' refcnt 0
  def swap_outfiles(name1, name2)
    do_command "outfile swap #{name1} #{name2}\n"
  end

  def close_outfile(name)
    do_command "outfile close #{name}\n"
  end

  # {parameters} should be a hash.  The following are the valid keys, which
  # may be symbols or strings (value types are S=string, I=int, B=bool):
  #
  #  S descr:    description of source
  #  I id:       source ID (list ID in skitter parlance)
  #  S file:     path of input file containing destination addresses
  #  I priority: probing chunk size (# addresses) used for scheduling
  #  B adhoc:    whether addresses will be fed via control socket
  #  S outfile:  name (not path) of output file
  #  I cycle:    # cycles through the destination list to probe if 'adhoc off'
  #  B autoreload: whether to re-load the destination list at each cycle
  #
  def add_source(name, parameters)
    command = "source add name \"#{name}\""
    parameters.each do |k,v|
      case k.to_s
      when "descr", "file", "outfile"
	command << " #{k} \"#{v}\""

      when "adhoc", "autoreload"
	command << " #{k} " << (v ? "on" : "off")

      else
	command << " #{k} #{v}"
      end
    end
    command << "\n"
    do_command command
  end

  def delete_source(name)
    do_command "source delete \"#{name}\"\n"
  end

  def cycle_source(name)
    do_command "source cycle \"#{name}\"\n"
  end

  def list_sources
    sock_send "source list\n"
    response = sock_read_response
    return nil if response[-1] =~ /^ERR/

    retval = []
    response.each_line do |line|
      next unless line =~ /^INFO (.+)$/
      retval << $1
    end
    retval
  end

  # Switches this control socket connection into source-event monitoring mode.
  #
  # Every time a source is added, removed, or starts a new cycle, an event
  # line of the form "EVENT <timestamp> <event-details>" is passed to the
  # block given to this method.  The block should return true to continue
  # monitoring, and false to halt monitoring and close the socket.  Note
  # that scamper does _not_ allow you to exit event monitoring mode and
  # return to regular command mode on a control socket--this is by design.
  # Thus, once this method returns, this Scamper::Control object will be
  # defunct.  To issue another command, another Scamper::Control object
  # must be created first.
  #
  # The syntax of event lines closely mimics the normal command syntax.
  # Example event lines are
  #
  # EVENT 1166121542 source add name 'CAIDA' id 1 file '/Users/youngh/caida-hosts' cycle 1 autoreload off priority 1 adhoc off outfile 'OUT'
  # EVENT 1166121542 source cycle 'CAIDA' id 1
  # EVENT 1166121543 source finish 'CAIDA'
  #
  # Note that scamper will always use single quotes around identifiers and
  # strings.  Any particular ordering of command parameters should not be
  # relied upon.
  #
  # The event line passed to the supplied block will not have a trailing
  # newline.
  def observe_sources
    sock_send "observe sources\n"
    line = @sock.gets
    raise_connection "read EOF" unless line
    raise_connection "'observe sources' failed" unless line == "OK\n"

    begin
      loop do
        line = @sock.gets
        raise_connection "read EOF" unless line

        yield line.chomp
      end
    ensure
      @sock.close rescue nil if @sock
      @sock = nil
    end
  end

  # {destination} should be a string giving an IPv4 or IPv6 address
  # Returns the name of the source to which the traceroute was added or nil
  # on error.
  def traceroute(destination, source=nil)
    command = "traceroute "
    command << "source \"#{source}\" " if source
    command << destination << "\n"
    sock_send command
    response = sock_read_response
    response[0] =~ /^OK .+ source '(.+)'$/
    $1
  end

  # note: "shutdown blah" => "ERR unhandled command 'blah'"
  def shutdown(how="done")
    begin
      command = "shutdown"
      command << " " << how if how
      command << "\n"
      sock_send command
      sock_read_response
      # scamper shuts down the socket unless {how} is invalid.
    rescue ConnectionBroken
      # ignore this to simplify clients who don't care about connectivity
    else
      @sock.close rescue nil if @sock
      @sock = nil
    end
  end

end

end  # module ArkUtil

