#############################################################################
## Various utility methods useful for working with Net::SSH and Net::SFTP.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: sshutils.rb,v 1.6 2009/05/22 22:52:52 youngh Exp $
#############################################################################

require 'arkutil/generalutils'
require 'arkutil/interactiveutils'

module ArkUtil

  # Returns true if the path is a directory on the remote system.
  # Returns nil if this can't tell, either because Net::SFTP can't tell
  # or the path doesn't exist 
  def self.sftp_remote_directory?(sftp, destination_path)
    begin
      attributes = sftp.stat!(destination_path)
      return attributes.directory?
    rescue Net::SFTP::StatusException
      return nil
    rescue
      raise
    end
  end


  def self.sftp_upload(sftp, local_path, remote_path)
    # Note: Net::SFTP v2.0.0 doesn't support a :preserve option for #upload,
    #       so we have to manually set the file attributes.
    local_stat = File.stat local_path

    sftp.upload!(local_path, remote_path) do
      |event, uploader, *args|
      case event
      when :open
        # args[0] : file metadata
        puts "uploading: #{args[0].local} -> #{args[0].remote}" +
          " (#{args[0].size} bytes)"

      when :put
        # args[0] : file metadata
        # args[1] : byte offset in remote file
        # args[2] : data being written (as string)
        if args[1] > 0
          local_file = File.basename(args[0].local)
          printf "  %2.0f%% (%d bytes): %s\r",
          100.0 * args[1].quo(args[0].size), args[1], local_file
          STDOUT.flush
        end

      when :close
        # args[0] : file metadata
        local_file = File.basename(args[0].local)
        puts "  finished #{local_file}                           "

      when :mkdir
        # args[0] : remote path name
        puts "  creating directory #{args[0]}"
        #when :finish
      end
    end

    # With Net::SFTP v2.0.0, we *must* set both :atime and :mtime together.
    # Otherwise, we'll trigger an exception because of the way
    # Net::SFTP::Protocol::V01::Attributes#to_s is implemented.
    sftp.setstat! remote_path, :permissions => (local_stat.mode & 07777),
    :atime => local_stat.mtime.to_i, :mtime => local_stat.mtime.to_i,
    :ctime => local_stat.ctime.to_i
  end


  # If the remote file doesn't exist, then this raises
  # Net::SFTP::StatusException
  # with code == Net::SFTP::Constants::StatusCodes::FX_NO_SUCH_FILE.
  def self.sftp_download(sftp, remote_path, local_path)
    # Note: Net::SFTP v2.0.0 doesn't support a :preserve option for #download,
    #       so we have to manually set the file attributes.
    attributes = sftp.stat! remote_path
    remote_size = attributes.size

    sftp.download!(remote_path, local_path) do |event, downloader, *args|
      case event
      when :open
        # args[0] : file metadata
        puts "downloading: #{args[0].remote} -> #{args[0].local}" +
          "(#{remote_size} bytes)"

      when :get
        # args[0] : file metadata
        # args[1] : byte offset in remote file
        # args[2] : data being written (as string)
        if args[1] > 0
          remote_file = File.basename(args[0].remote)
          printf "  %2.0f%% (%d bytes): %s\r",
          100.0 * args[1].quo(remote_size), args[1], remote_file
          STDOUT.flush
        end

      when :close
        # args[0] : file metadata
        remote_file = File.basename(args[0].remote)
        puts "  finished #{remote_file}                           "

      when :mkdir
        # args[0] : local path name
        puts "  creating directory #{args[0]}"

        #when :finish
      end
    end

    File.chmod attributes.permissions & 07777, local_path
    File.utime attributes.atime, attributes.mtime, local_path
  end


  def self.upload_file(host, port, user, local_path, remote_path)
    Net::SFTP.start(host, user, :port => port) do |sftp|
      begin
        sftp_upload sftp, local_path, remote_path
      rescue Net::SFTP::StatusException => e
        $stderr.puts "error: #{local_path} -> #{remote_path}: " +
          Net::SFTP::Response::MAP[e.code]
        raise
      rescue
        $stderr.puts "error: #{local_path} -> #{remote_path}: " +
          $!.inspect
        raise
      end
    end
  end


  def self.download_file(host, port, user, remote_path, local_path)
    Net::SFTP.start(host, user, :port => port) do |sftp|
      begin
        sftp_download sftp, remote_path, local_path
      rescue Net::SFTP::StatusException => e
        $stderr.puts "error: #{remote_path} -> #{local_path}: " +
          Net::SFTP::Response::MAP[e.code]
        raise
      rescue
        $stderr.puts "error: #{remote_path} -> #{local_path}: " +
          $!.inspect
        raise
      end
    end
  end


#############################################################################
# This class is useful for executing shell commands.
#
# This class knows about sudo and will transparently cache the sudo password
# to re-use across ssh executions (even across monitors).
class SSHExecHelper

  def initialize(abort_on_exec_error=false)
    @exec_summary = Hash.new 0  # :exit_error, etc. => count
    @abort_on_exec_error = abort_on_exec_error
  end

  # Returns :exec_failed, :exit_error, :exit_success, or :signal.
  def exec(ssh, command)
    puts "[executing: " + command + "]"
    $stdout.flush

    retval = nil
    status = "UNKNOWN STATUS"
    @password_prompt_count = 0
    ssh.open_channel do |channel|
      channel.exec(command) do |ch, success|
        unless success
          retval = :exec_failed
          status = "FAILED: couldn't execute command (ssh.channel.exec failure)"
        end

        channel.on_data do |ch, data|  # stdout
          print data
        end

        channel.on_extended_data do |ch, type, data|
          next unless type == 1  # only handle stderr
          $stderr.print data
          channel[:output] ||= ""
          channel[:output] += data
          if channel[:output] =~ /^Password:/ # ^ can match interior line start
            channel[:output] = ""
            channel.send_data password() + "\n"
          elsif channel[:output] =~ /^sudo: \d incorrect password attempt/
            # To continue would trigger sudo password failures across remaining
            # monitors, so abort.
            abort "sudo failed!"
          end
        end

        channel.on_request("exit-status") do |ch, data|
          exit_code = data.read_long
          if exit_code > 0
            retval = :exit_error
            status = "ERROR: exit code #{exit_code}"
          else
            retval = :exit_success
            status = "success"
          end
        end

        channel.on_request("exit-signal") do |ch, data|
          retval = :signal
          status = "SIGNAL: #{data.read_long}"
        end
      end
    end
    ssh.loop

    puts "[finished: #{status}]"
    $stdout.flush

    if retval != :exit_success && @abort_on_exec_error
      abort "Exiting on first ssh exec failure."
    end

    @exec_summary[retval] += 1
    retval
  end

  def password
    if @password == nil || @password_prompt_count > 0
      @password = get_password()
    end
    @password_prompt_count += 1
    @password
  end

  def success_count
    @exec_summary[:exit_success]
  end

  def fail_count
    exec_count = ArkUtil.sum(@exec_summary.values)
    success_count = @exec_summary[:exit_success]
    exec_count - success_count
  end

  def report_exec_summary
    exec_count = ArkUtil.sum(@exec_summary.values)
    success_count = @exec_summary[:exit_success]
    fail_count = exec_count - success_count

    return unless exec_count > 1  # reduce verbiage
    printf "\nExecution summary: %s, %s\n",
      ArkUtil.count_with_unit(success_count, "success", "es"),
      ArkUtil.count_with_unit(fail_count, "failure", "s")

    [:exit_error, :signal, :exec_failed].each do |key|
      count = @exec_summary[key]
      next if count == 0
      print "   >> #{count} "
      case key
      when :exit_error then puts "had non-zero exit status"
      when :signal then puts "exited on a signal"
      when :exec_failed then puts "could not be executed"
      end
    end
  end


  private # .............................................................

  def get_password()
    ArkUtil.disable_echo do
      input = $stdin.gets
      if input == nil
        abort "Aborting sudo."  # /usr/bin/sudo also exits on ^D at Password:
      end
      return input.chomp
    end
  end
end

end  # module ArkUtil
