#############################################################################
## This class provides a wrapper around tmux, the terminal multiplexer.
##
## --------------------------------------------------------------------------
## Copyright (C) 2012 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: tmux.rb,v 1.8 2016/07/18 22:43:06 youngh Exp $
#############################################################################

require 'open4'

module ArkUtil

class Tmux

  class NoServerError < StandardError; end

  class ExecutionError < StandardError
    attr_reader :status, :stdout, :stderr

    def initialize(status, stdout, stderr)
      super()
      @status, @stdout, @stderr = status, stdout, stderr
    end

    def to_s
      sprintf "%p; stdout=%p; stderr=%p", status, stdout, stderr
    end
  end

  @@default_options = {}

  # A given Tmux instance is bound to a particular server socket and session,
  # so a single Tmux instance cannot be used with multiple sessions.  To use
  # multiple sessions, create separate Tmux instances for each distinct
  # session.
  #
  # Supported options:
  #
  #   :debug => bool
  #   :server_socket => path  # path to the server socket
  #   :session => name
  def initialize(options={})
    @options = @@default_options.dup
    @options.merge! options
    @debug = @options[:debug]

    @tmux = "tmux"
    if @options[:server_socket]
      @tmux += " -S \"" + @options[:server_socket] + "\""
    end

    @session = @options[:session]
  end


  def server_running?
    begin
      status, stdout, stderr = execute "list-session"
      if status.success?
        return true
      else
        raise ExecutionError.new(status, stdout, stderr)
      end
    rescue NoServerError
      return false
    end
  end


  # Starts the tmux server, if not already running, and creates a new session
  # with the name specified in the options passed to Tmux.new.
  #
  # Note: There's no way to start a new tmux session without also
  #       creating a new window, which is why this takes a {window}
  #       to name the first created window.
  #
  # This raises ExecutionError on any error, including when the session already
  # exists.
  #
  # Impl Note: tmux outputs "duplicate session: ..." and exits with code 1.
  def new_session(window=nil)
    arguments = "new-session -d"
    arguments += sprintf " -s \"%s\"", @session if @session
    arguments += sprintf " -n \"%s\"", window if window

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Kills the session with the name specified in the options passed to Tmux.new.
  #
  # Returns true if the session was killed and false if no matching session
  # exists.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  #
  # Impl Note: If the session doesn't exist, tmux outputs "session not
  #            found: ..." and exits with code 1.
  def kill_session
    unless @session
      raise ArgumentError, "no configured session name"
    end

    arguments = sprintf "kill-session -t \"%s\"", @session

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    elsif stderr.include? "session not found"
      return false
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Returns an array of session names.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def list_sessions
    # Be careful: tmux uses the Ruby substitution syntax "#{ }" for -F.
    arguments = 'list-sessions -F "#{session_name}"'

    status, stdout, stderr = execute arguments
    if status.success?
      return stdout.split "\n"
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Returns the index of the new window.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  #
  # Note: The -t argument must end with a colon to avoid an obscure ambiguity
  #       caused by the target session having a name that is equal to or a
  #       prefix of an existing window, which will cause new-window to fail.
  #       For example, suppose the session is called "foo" and there exists
  #       a window named "foo-bar" at index 3.  Then
  #       
  #        $ tmux new-window -P  -n "abc" -t "foo"
  #        create window failed: index in use: 3
  #
  #       The new-window command handles -t as follows (from tmux(1)):
  #
  #          When the [-t] argument does not contain a colon, tmux
  #          first attempts to parse it as window; if that fails, an
  #          attempt is made to match a session....
  #
  #          'target-window' represents the window to be created; if
  #          the target already exists an error is shown, unless the
  #          -k flag is used, in which case it is destroyed.
  def new_window(name=nil)
    # -P prints out the index of the new window as "<session>:<index>"
    arguments = "new-window -P "
    arguments += sprintf " -n \"%s\"", name if name
    arguments += sprintf " -t \"%s:\"", @session if @session

    status, stdout, stderr = execute arguments
    if status.success?
      if stdout =~ /\:(\d+)(\.\d+)?$/
        return $1.to_i
      else
        fail "INTERNAL ERROR: couldn't parse the output of 'tmux new-window'"
      end
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Kills a window specified by either a name or an index (an index is
  # preferred, since a name may not be unique).
  #
  # Returns true if the window was killed and false if no matching window
  # exists.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  #
  # Impl Note: If the window doesn't exist, tmux outputs "window not found: ..."
  #            and exits with code 1.
  def kill_window(name=nil, index=nil)
    id = (index || name)
    unless id
      raise ArgumentError, "must provide either 'name' or 'index'"
    end

    arguments = "kill-window"
    target = ""
    target += @session + ":" if @session
    target += id.to_s
    arguments += sprintf " -t \"%s\"", target

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    elsif stderr.include? "window not found"
      return false
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Returns [ [ index, window_name ] ] (for the configured session).
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def list_windows
    # Be careful: tmux uses the Ruby substitution syntax "#{ }" for -F.
    arguments = 'list-windows -F "#{window_index} #{window_name}"'
    arguments += sprintf " -t \"%s\"", @session if @session

    status, stdout, stderr = execute arguments
    if status.success?
      retval = []
      stdout.each_line do |line|
        line.chomp!
        index, name = line.split(" ", 2)
        retval << [index.to_i, name]
      end
      return retval
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def load_buffer_from_file(path)
    arguments = sprintf "load-buffer \"%s\"", path

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def load_buffer_from_string(s)
    arguments = "load-buffer -"

    status, stdout, stderr = execute arguments, s
    if status.success?
      return true
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Pastes the current buffer to a window specified by either a name
  # or an index (an index is preferred, since a name may not be
  # unique).
  #
  # Returns true on success and false if no matching window exists.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  #
  # Impl Note: If the window doesn't exist, tmux outputs "window not found: ..."
  #            or "session not found: ..." and exits with code 1.  
  def paste_buffer(name=nil, index=nil)
    id = (index || name)
    unless id
      raise ArgumentError, "must provide either 'name' or 'index'"
    end

    arguments = "paste-buffer"
    target = ""
    target += @session + ":" if @session
    target += id.to_s
    arguments += sprintf " -t \"%s\"", target

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    elsif stderr.include?("not found")
      return false
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Saves a buffer to a file, optionally appending to an existing file
  # (if {append} is true).
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def save_buffer(path, append=false)
    arguments = "save-buffer"
    arguments += " -a" if append
    arguments += " \"" + path + "\""

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Returns the contents of the topmost buffer or the buffer at the
  # specified index.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def show_buffer(index=nil)
    arguments = "show-buffer"
    arguments += sprintf " -b %d", index if index

    status, stdout, stderr = execute arguments
    if status.success?
      return stdout
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Deletes the topmost buffer or the buffer at the specified index.
  #
  # This is useful for freeing up large buffers created with capture_pane.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def delete_buffer(index=nil)
    arguments = "delete-buffer"
    arguments += sprintf " -b %d", index if index

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Deletes all buffers.
  #
  # This is useful for freeing up large buffers created with capture_pane.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  def delete_all_buffers
    arguments = "list-buffers"

    status, stdout, stderr = execute arguments
    unless status.success?
      raise ExecutionError.new(status, stdout, stderr)
    end

    # Buffer indexes are dynamic.  When a buffer at a given index is deleted,
    # all buffers with higher indexes shift down one in index.  Hence, don't
    # bother deleting buffers by index (unless you do so from high index to
    # low).
    stdout.each_line do |line|
      # 0: 19 bytes: "echo `date`\012whoami\012"
      delete_buffer() if line =~ /^\d+:/
    end
  end


  # Captures the scrollback of a window/pane into a buffer.
  # 
  # Afterwards, call save_buffer to save the buffer to a file.
  #
  # The {start} argument specifies the index of the starting line to capture.
  # The first visible line in a pane has index 0, and lines below that have
  # positive indexes and lines above have negative indexes.  There is no
  # harm in specifying a starting line number that goes farther back than
  # available in the scrollback (which is why the default starting value
  # is -10_000).
  #
  # Returns true on success and false if no matching window exists.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  #
  # Impl Note: If the window doesn't exist, tmux outputs "window not found: ..."
  #            or "session not found: ..." and exits with code 1.  
  def capture_pane(name=nil, index=nil, start=-10_000)
    id = (index || name)
    unless id
      raise ArgumentError, "must provide either 'name' or 'index'"
    end

    arguments = "capture-pane"
    target = ""
    target += @session + ":" if @session
    target += id.to_s
    arguments += sprintf " -t \"%s\" -S %d", target, start

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    elsif stderr.include?("not found")
      return false
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  # Clears the scrollback of the window specified by either a name
  # or an index (an index is preferred, since a name may not be
  # unique).
  #
  # Returns true on success and false if no matching window exists.
  #
  # This raises NoServerError if the tmux server isn't running.
  # This raises ExecutionError on any other error.
  #
  # Impl Note: If the window doesn't exist, tmux outputs "window not found: ..."
  #            or "session not found: ..." and exits with code 1.  
  def clear_history(name=nil, index=nil)
    id = (index || name)
    unless id
      raise ArgumentError, "must provide either 'name' or 'index'"
    end

    arguments = "clear-history"
    target = ""
    target += @session + ":" if @session
    target += id.to_s
    arguments += sprintf " -t \"%s\"", target

    status, stdout, stderr = execute arguments
    if status.success?
      return true
    elsif stderr.include?("not found")
      return false
    else
      raise ExecutionError.new(status, stdout, stderr)
    end
  end


  #--------------------------------------------------------------------------

  # Executes a raw tmux command.
  # This raises NoServerError if the tmux server isn't running.
  def execute(tmux_arguments, stdin=nil)
    # holds output from the executed command
    stdout = stderr = nil
    begin
      spawn_options = {}
      spawn_options['stdin'] = stdin if stdin
      spawn_options['stdout'] = stdout = ""
      spawn_options['stderr'] = stderr = ""
      spawn_options['ignore_exit_failure'] = true

      command = @tmux + " " + tmux_arguments
      Open4.spawn command, spawn_options
      status = extract_exit_status()

      if @debug
        $stderr.printf "Tmux#execute(%p)\n", command
        $stderr.printf "  stdout>> %p\n", stdout
        $stderr.printf "  stderr>> %p\n", stderr
        $stderr.printf "  status>> %p\n", status
      end

    rescue StandardError => exn
      status = [ "exception", exn.inspect + " at " + exn.backtrace.join(" <= ")]
    end

    class <<status
      def success?
        return self[0] == :exited && self[1] == 0
      end
    end

    if status[0] == :exited && status[1] != 0 &&
        stderr.include?("failed to connect to server")
      raise NoServerError
    end

    return status, stdout, stderr
  end


  private #-----------------------------------------------------------------

  def extract_exit_status
    retval = []
    if $?.exited?
      retval.push :exited, $?.exitstatus
    elsif $?.signaled?
      retval.push :signaled, $?.termsig
    elsif $?.stopped?
      retval.push :stopped, $?.stopsig
    # elsif $?.coredump?
      # Don't return info about $?.coredump?, since that only provides
      # additional information for the $?.signaled? case.
    end
    retval
  end

end

end  # module ArkUtil
